# SPDX-FileCopyrightText: (c) 2020 Art Galkin <ortemeo@gmail.com>
# SPDX-License-Identifier: BSD-3-Clause

import sys
import datetime as dt
from typing import *
from pathlib import Path

from depz import __version__
from depz.x00_common import Mode, printVerbose
from depz.x98_dooo import doo, OutputMode

helptxt = """

DEPZ is a simple dependency management utility. It allows to keep 
reusable code in local directories. Without packaging it as a library 
for distribution. 

The dependences of the project must be specified in text file 
named "depz.txt" in the project root dir.

Each line of "depz.txt" specifies a dependency.

Sample lines:

  "/abs/path/mylib"
  "../mylib"
  "~/path/mylib"
    The project depends on library "mylib" located in LOCAL dir
  
  "requests"
    The project depends on some EXTERNAL package called "requests"

When the project depends on local `mylib`, it means, it also depends on all 
the dependencies of `mylib`. So after scannig "project/depz.txt" we will also 
scan "mylib/depz.txt" the same way.

Local dependences are usually linked or copied into the project dir. 
For example, "/path/to/mylib" will lead to creation of symlink "./mylib".

External dependences are installed with tools like pip.

Exact behavior is specified by the program arguments.

"""


def runmain(programArgs: List[str]):
	import argparse

	parser = argparse.ArgumentParser()

	parser.add_argument("-p", "--project", type=str, default=".",
						help='The project directory path. Defaults to the current directory (".")')
	parser.add_argument("-m", "--mode", type=str, default="default", choices=["default", "layout"],
						help='The link creation mode. See docs.')

	parser.add_argument("-e", type=str, default="default", choices=["default", "line", "multi"],
						help='When specified, only external dependencies will be printed to stdout.')

	parser.add_argument("--relink", action="store_true",
						help="Remove all symlinks from the project dir and create symlinks to local dependencies")

	parser.add_argument("--version", action="store_true",
						help="Print version and exit")

	args = parser.parse_args(programArgs)

	if args.version:
		lastMod = dt.datetime.fromtimestamp(
			(Path(__file__).parent / "x00_version.py").stat().st_mtime)
		print(f"DEPZ {__version__} (c) 2020-{lastMod.year} Art Galkin <ortemeo@gmail.com>")
		exit(0)

	mode: Mode
	if args.mode == "default":
		mode = Mode.default
	elif args.mode == "layout":
		mode = Mode.layout
	else:
		raise ValueError

	outputMode: OutputMode

	if args.e == "default":
		outputMode = OutputMode.default
	elif args.e == "line":
		printVerbose.allowed = False
		outputMode = OutputMode.one_line
	elif args.e == "multi":
		printVerbose.allowed = False
		outputMode = OutputMode.multi_line
	else:
		raise ValueError

	doo(Path(args.project),
		symlinkLocalDeps=True,
		mode=mode, outputMode=outputMode)


if __name__ == "__main__":
	runmain(sys.argv[1:])



# SPDX-FileCopyrightText: (c) 2020 Art Galkin <ortemeo@gmail.com>
# SPDX-License-Identifier: BSD-3-Clause


from collections import deque, defaultdict
from pathlib import Path
from typing import *

from depz.x00_common import Mode, printVerbose
from depz.x01_testsBase import TestWithDataDir
from depz.x50_resolve import resolvePath
from depz.x50_unlink import unlinkAll


def pathToLibname(path: Path) -> str:
	n = path.name
	if n.endswith("_py"):
		n = n[:-3]
	elif n.endswith("_flutter"):
		n = n[:-8]

	return n


def _debugIterParents(p: Path) -> Iterator[Path]:
	"""Returns /path/to/parent/file, /path/to/parent, /path/to, /path, /"""
	parts = list(p.parts)
	for l in range(len(parts), 0, -1):
		yield Path(*parts[:l])


def symlinkVerbose(realPath: Path, linkPath: Path, targetIsDirectory: bool,
				   createLinkParent: bool = False):
	"""Creates a symlink. Throws more detailed exceptions, than Path.

	Path.symlink_to can throw a FileNotFound error without making it clear what is missing:
	source or target.
	"""

	if not realPath.exists():
		for par in _debugIterParents(realPath):
			printVerbose(par, par.exists())
		raise FileNotFoundError(f"realPath path {realPath} does not exist")
	if createLinkParent:
		linkPath.parent.mkdir(parents=True, exist_ok=True)
	elif not linkPath.parent.exists():
		raise FileNotFoundError(f"The parent dir of destination linkPath {linkPath} does not exist")

	printVerbose("Creating symlink:")
	printVerbose(f"  src: {realPath.absolute()}")
	printVerbose(f"  dst: {linkPath.absolute()}")
	linkPath.symlink_to(realPath, target_is_directory=targetIsDirectory)


def symlinkPython(srcLibDir: Path, dstPythonpathDir: Path):
	# создает в каталоге dstPythonpathDir ссылку на библиотеку, расположенную в srcLibDir

	name = pathToLibname(srcLibDir)
	symlinkVerbose(srcLibDir, dstPythonpathDir / name, targetIsDirectory=True)


def symlinkLayout(srcLibDir: Path, dstProjectDir: Path):
	"""Creates symlinks from items inside srcLibDir to items inside dstProjectDir

	libraryA/lib	-> project/lib/libraryA
	libraryA/test	-> project/test/libraryA

	libraryB/lib	-> project/lib/libraryB
	libraryB/test	-> project/test/libraryB

	"""

	libName = pathToLibname(srcLibDir)

	for item in srcLibDir.glob("*"):
		if item.is_dir():
			symlinkVerbose((srcLibDir / item.name).absolute(), dstProjectDir / item.name / libName,
						   targetIsDirectory=True, createLinkParent=True)


def pydpnFiles(dirPath: Path) -> Iterable[Path]:
	for p in [
		dirPath / "depz.txt",
		dirPath / "lib" / "depz.txt",
		dirPath / "pydpn.txt",  # deprecated since 2021-03
		dirPath / "lib" / "pydpn.txt"  # deprecated since 2021-03
	]:
		if p.exists():
			yield p


def iterLnkdpnLines(file: Path) -> Iterator[str]:
	"""Returns all lines except empty and comments"""
	for line in file.read_text().splitlines():
		line = line.partition("#")[0].strip()
		if line:
			yield line


def rescan(projectDir: Path, relink: bool, mode: Mode) -> Dict[str, Set[str]]:
	# сканирует файл depz.txt в каталоге проекта, а также, следуя по ссылкам на другие локальные
	# библиотеки - все файлы pydpn.txt в тех библиотеках.
	#
	# В итоге на локальные библиотеки делает симлинки в каталоге project/pkgs (заменяя все
	# симлинки, которые там были).
	#
	# А имена внешних библиотек просто возвращает списокои

	localLibs: Set[Path] = set()
	externalLibs = defaultdict(set)

	# читаю все pydpn, запоминая результаты, но ничего не меняя

	pathsToAnalyze: Deque[Path] = deque((projectDir.absolute(),))

	while pathsToAnalyze:

		currDir = pathsToAnalyze.popleft()  # каталог проекта или библиотеки

		for lnkdpnFile in pydpnFiles(currDir):

			printVerbose(f"Depz file: {lnkdpnFile}")

			for line in iterLnkdpnLines(lnkdpnFile):

				localPkgPath = resolvePath(lnkdpnFile.parent, line)

				if localPkgPath:

					localPkgPath = localPkgPath.absolute()
					if not localPkgPath in localLibs:
						localLibs.add(localPkgPath)
						pathsToAnalyze.append(localPkgPath)

				else:

					assert not localPkgPath
					externalLibs[line].add(pathToLibname(currDir))

	if relink:

		# пересоздаю симлинки (удаляю все, создаю актуальные)

		pkgsDir = projectDir  # /PKGSBN
		pkgsDir.mkdir(exist_ok=True)
		unlinkAll(pkgsDir)
		if mode == Mode.layout:
			# if isFlutterDir(pkgsDir):
			unlinkAll(pkgsDir / "lib")
			unlinkAll(pkgsDir / "test")

		# isFlutter = isFlutterDir(projectDir)

		# printVerbose("localLibs:", localLibs)

		for path in localLibs:
			if mode == Mode.layout:
				symlinkLayout(path, pkgsDir)
			else:
				symlinkPython(path, pkgsDir)

	# возвращаю то, что не было ссылками на локальные проекты: т.е. внешние зависимости

	return externalLibs



from enum import IntEnum, auto
from pathlib import Path

from depz.x00_common import Mode, printVerbose
from depz.x80_rescanRelink import rescan


class OutputMode(IntEnum):
	default = auto()
	one_line = auto()
	multi_line = auto()


def doo(projectPath: Path,
		symlinkLocalDeps: bool = False,
		mode: Mode = Mode.default,
		outputMode: OutputMode = OutputMode.default):
	printVerbose(f"Project dir: {projectPath.absolute()}")

	externalLibs = rescan(projectPath, relink=symlinkLocalDeps, mode=mode)

	if outputMode == OutputMode.default:
		if externalLibs:
			print("External dependencies:", " ".join(externalLibs))
		else:
			print("No external dependencies.")
	elif outputMode == OutputMode.one_line:
		print(" ".join(externalLibs))
	elif outputMode == OutputMode.multi_line:
		print("\n".join(externalLibs))
	else:
		raise ValueError


# SPDX-FileCopyrightText: (c) 2021 Art Galkin <ortemeo@gmail.com>
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations
import os
import sys
import unittest
from io import StringIO
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import List, Iterator, Tuple

from depz import runmain


class CapturedOutput:
	# ? maybe replace with https://pypi.org/project/stream-redirect/
	def __init__(self):
		self._new_out = StringIO()
		self._new_err = StringIO()

	def __enter__(self) -> CapturedOutput:
		self._old_stdout = sys.stdout
		self._old_stderr = sys.stderr
		sys.stdout = self._new_out
		sys.stderr = self._new_err
		return self

	def __exit__(self, exc_type, exc_val, exc_tb):
		sys.stdout = self._old_stdout
		sys.stderr = self._old_stderr

	@property
	def std(self) -> str:
		return self._new_out.getvalue()

	@property
	def err(self) -> str:
		return self._new_err.getvalue()


def rglobWithSymlinks(parent: Path, mask="*") -> Iterator[Path]:
	# because in 2020 Path.rglob("*") does not follow symlinks
	# https://bugs.python.org/issue33428

	for p in parent.rglob(mask):
		yield p
		if p.is_symlink():
			for sub in rglobWithSymlinks(p, mask):
				yield sub


def listDir(dir: Path, withFileSizes=False):
	result: List[str] = list()

	for p in rglobWithSymlinks(dir):
		kind = ""
		size = ""
		if p.is_symlink():
			kind += "L"
		if p.is_dir():
			kind += "D"
		elif p.is_file():
			kind += "F"
			if withFileSizes:
				size = " " + str(p.lstat().st_size)
		else:
			raise ValueError(f"Unexpected entry: {p} {p.lstat().st_mode}")

		result.append(f"{p.relative_to(dir)} ({kind}{size})")

	result.sort()

	return result


def createFile(path: Path, content: str = None):
	path.parent.mkdir(parents=True, exist_ok=True)
	if content:
		path.write_text(content)
	else:
		path.touch()


class Tests(unittest.TestCase):
	def setUp(self) -> None:
		self.td = TemporaryDirectory()
		self.tempDir = Path(self.td.name)
		self.createLayout()

	def tearDown(self) -> None:
		self.td.cleanup()

	def createLayout(self):
		raise NotImplementedError


class TestsWithPythonLayout(Tests):

	def createLayout(self):
		createFile(self.tempDir / "project" / "stub.py")
		createFile(self.tempDir / "project" / "depz.txt",
				   """	# local
					   ../libs/lib1
					   ../libs/lib2
					   # external
					   numpy 
				   """)

		createFile(self.tempDir / "libs" / "__init__.py")
		createFile(self.tempDir / "libs" / "lib1" / "depz.txt",
				   """	../lib3
					   requests
				   """)
		createFile(self.tempDir / "libs" / "lib2" / "__init__.py")
		createFile(self.tempDir / "libs" / "lib3" / "__init__.py")

	expectedPythonAfterLink = [
		'depz.txt (F)',
		'lib1 (LD)',
		'lib1/depz.txt (F)',
		'lib2 (LD)',
		'lib2/__init__.py (F)',
		'lib3 (LD)',
		'lib3/__init__.py (F)',
		'stub.py (F)']

	def test_default_mode(self):
		runmain(["--project", str(self.tempDir / "project"), "--relink"])
		result = listDir((self.tempDir / "project"))
		self.assertListEqual(result, self.expectedPythonAfterLink)

	def _run_relink_current_dir(self):
		runmain(["--relink"])
		result = listDir((self.tempDir / "project"))
		self.assertListEqual(result, self.expectedPythonAfterLink)

	def test_current_dir_as_project_dir(self):
		# when project dir not specified, use the current dir

		os.chdir(str(self.tempDir / "project"))  # changing current dir the project
		self._run_relink_current_dir()

	@unittest.expectedFailure
	def test_project_dir_error(self):
		# running in the wrong directory without specifying --project

		wrongDir = self.tempDir / "wrong"
		wrongDir.mkdir()
		os.chdir(str(wrongDir))

		self._run_relink_current_dir()

	def test_print_externals_one_line(self):
		with CapturedOutput() as output:
			runmain(["--project", str(self.tempDir / "project"), "--relink", "-e", "line"])
		self.assertEqual(output.std.strip(), "numpy requests")

	def test_print_externals_multi_lines(self):
		with CapturedOutput() as output:
			runmain(["--project", str(self.tempDir / "project"), "--relink", "-e", "multi"])
		self.assertEqual(output.std.strip(), "numpy\nrequests")


class TestsWithFluterLayout(Tests):

	def createLayout(self):
		(self.tempDir / "project" / "lib").mkdir(parents=True)
		(self.tempDir / "project" / "test").mkdir(parents=True)

		createFile(self.tempDir / "project" / "depz.txt",
				   """
					../libraryA_flutter
					../libraryB
					externalLib
				   """)

		createFile(self.tempDir / "project" / "pubspec.yaml")

		createFile(self.tempDir / "libraryA_flutter" / "pubspec.yaml")  # must not be symlinked
		createFile(self.tempDir / "libraryA_flutter" / "lib" / "code1.dart")
		createFile(self.tempDir / "libraryA_flutter" / "lib" / "code2.dart")
		createFile(self.tempDir / "libraryA_flutter" / "test" / "testA.dart")

		createFile(self.tempDir / "libraryB" / "lib" / "code3.dart")
		createFile(self.tempDir / "libraryB" / "test" / "testB.dart")
		createFile(self.tempDir / "libraryB" / "pydpn.txt",
				   """
					   ../libraryC
					   externalFromC
				   """)

		createFile(self.tempDir / "libraryC" / "lib" / "something.dart")
		createFile(self.tempDir / "libraryC" / "data" / "binary.dat")

	expectedAfterLink = ['data (D)',
						 'data/libraryC (LD)',
						 'data/libraryC/binary.dat (F)',
						 'depz.txt (F)',
						 'lib (D)',
						 'lib/libraryA (LD)',
						 'lib/libraryA/code1.dart (F)',
						 'lib/libraryA/code2.dart (F)',
						 'lib/libraryB (LD)',
						 'lib/libraryB/code3.dart (F)',
						 'lib/libraryC (LD)',
						 'lib/libraryC/something.dart (F)',
						 'pubspec.yaml (F)',
						 'test (D)',
						 'test/libraryA (LD)',
						 'test/libraryA/testA.dart (F)',
						 'test/libraryB (LD)',
						 'test/libraryB/testB.dart (F)']

	def test_relink_layout(self):
		runmain(["--project", str(self.tempDir / "project"), "--relink", "--mode", "layout"])
		result = listDir((self.tempDir / "project"))

		from pprint import pprint

		self.assertListEqual(result, self.expectedAfterLink)


class TestsInfo(unittest.TestCase):
	def test_help(self):
		with CapturedOutput() as output:
			with self.assertRaises(SystemExit) as cm:
				runmain(["--help"])
		self.assertEqual(cm.exception.code, 0)
		self.assertTrue("arguments" in output.std)
		self.assertEqual(output.err, "")

	def test_version(self):
		with CapturedOutput() as output:
			with self.assertRaises(SystemExit) as cm:
				runmain(["--version"])
		self.assertEqual(cm.exception.code, 0)
		self.assertTrue(output.std.startswith("DEPZ"))
		self.assertEqual(output.err, "")



# SPDX-FileCopyrightText: (c) 2020 Art Galkin <ortemeo@gmail.com>
# SPDX-License-Identifier: BSD-3-Clause

import unittest
from pathlib import Path


class TestWithDataDir(unittest.TestCase):
	@property
	def dataDir(self) -> Path:
		d = Path(__file__).parent / "test" / "data"
		if not d.exists():
			raise FileNotFoundError(d)
		return d


class TestWithTempDir(unittest.TestCase):

	def setUp(self):
		from tempfile import TemporaryDirectory
		self._td = TemporaryDirectory()
		self.tempDir = Path(self._td.name)

	def tearDown(self) -> None:
		self._td.cleanup()

	@staticmethod
	def mkd(p: Path) -> Path:
		p.mkdir(parents=True)
		return p



# SPDX-FileCopyrightText: (c) 2020 Art Galkin <ortemeo@gmail.com>
# SPDX-License-Identifier: BSD-3-Clause


from enum import IntEnum, auto


class Mode(IntEnum):
	default = auto()
	layout = auto()


def printVerbose(text: str):
	if printVerbose.allowed:
		print(text)


printVerbose.allowed = True



# SPDX-FileCopyrightText: (c) 2020 Art Galkin <ortemeo@gmail.com>
# SPDX-License-Identifier: BSD-3-Clause

import os
from pathlib import Path
from typing import Optional

from depz.x01_testsBase import TestWithTempDir


def resolvePath(rootDir: Path, packageDir: str) -> Optional[Path]:
	"""Gives interpretation to a single line of depz.txt.

	:param rootDir: The directory where depz.txt found.
	:param packageDir: Either relative to the rootDir or absolute.
	:return: The path to the library directory. NULL, if directory does not exist
	or does not contain a library.
	"""

	packageDir = packageDir.strip()
	packageDir = os.path.expanduser(packageDir)
	packageDir = os.path.expandvars(packageDir)
	packageDir = os.path.normpath(packageDir)

	if os.path.isabs(packageDir):
		packageDirPath = Path(packageDir)
	else:
		packageDirPath = rootDir / packageDir
	packageDirPath = packageDirPath.resolve()

	if packageDirPath.exists() and packageDirPath.is_dir():
		return packageDirPath


class TestResolvePath(TestWithTempDir):

	def test_relative(self):
		projectDir = self.mkd(self.tempDir/"prj"/"project")
		libDir = self.mkd(self.tempDir/"libs"/"libA")
		# finding libDir by relative path
		self.assertTrue(libDir.samefile(resolvePath(projectDir, "../../libs/libA")))

	def test_absolute(self):
		projectDir = self.mkd(self.tempDir/"prj"/"project")
		libDir = self.mkd(self.tempDir/"libs"/"libA")
		# finding libDir by absolute path
		self.assertTrue(libDir.samefile(resolvePath(projectDir, str(libDir.absolute()))))

	def test_no_such_dir(self):
		projectDir = self.mkd(self.tempDir/"prj"/"project")
		self.assertEqual(resolvePath(projectDir, "linking_nowhere_2412648263486"), None)


from importlib.machinery import SourceFileLoader
from pathlib import Path

from setuptools import setup, find_packages

ver = SourceFileLoader('ver', 'depz/x00_version.py').load_module()

setup(
  name="depz",
  version=ver.__version__,

  author="Art Galkin",
  author_email="ortemeo@gmail.com",
  url='https://github.com/rtmigo/depz',

  packages=find_packages(),
  install_requires=[],

  description="Command line tool for symlinking directories with reusable code into the project",

  long_description=(Path(__file__).parent / 'README.md').read_text(),
  long_description_content_type='text/markdown',

  license='BSD-3-Clause',

  entry_points={
    'console_scripts': [
      'depz = depz:runmain',
    ]},

  keywords="""
  	source-code dependencies filesystem programming
  	""".split(),

  # https://pypi.org/classifiers/
  classifiers=[
    # "Development Status :: 4 - Beta",
    "Development Status :: 2 - Pre-Alpha",
    # "Development Status :: 3 - Alpha",
    "Intended Audience :: Developers",
    'License :: OSI Approved :: BSD License',
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Typing :: Typed",
    "Topic :: Software Development :: Build Tools",
    "Operating System :: POSIX"
  ],

  # test_suite='nose.collector',
  # tests_require=['nose'],
  #zip_safe=False
)


# SPDX-FileCopyrightText: (c) 2020 Art Galkin <ortemeo@gmail.com>
# SPDX-License-Identifier: BSD-3-Clause

from pathlib import Path

from depz.x01_testsBase import TestWithTempDir


def unlinkAll(parent: Path):
	# removes all symlinks that are immediate children of parent dir
	for child in parent.glob("*"):
		if child.is_symlink():
			child.unlink()


class TestUnlink(TestWithTempDir):
	def test(self):

		tempSubdir = (self.tempDir / "subdir" / "iLikeToBeLinkedTo").absolute()
		tempSubdir.mkdir(exist_ok=True, parents=True)

		link1 = (self.tempDir / "link1").absolute()
		link2 = (self.tempDir / "link2").absolute()

		if not link1.exists():
			link1.symlink_to(tempSubdir, True)

		if not link2.exists():
			link2.symlink_to(tempSubdir, True)

		self.assertTrue(link1.exists() and link1.is_symlink())
		self.assertTrue(link2.exists() and link2.is_symlink())

		unlinkAll(self.tempDir)

		self.assertFalse(link1.exists())
		self.assertFalse(link2.exists())
